import {GetConnectionCommand, GoneException, PostToConnectionCommand} from '@aws-sdk/client-apigatewaymanagementapi'
import {type AnyTRPCRouter, callTRPCProcedure, type inferRouterContext} from '@trpc/server'
import type {APIGatewayProxyEvent, APIGatewayProxyResult, Context} from 'aws-lambda'
import {getApiGatewayManagementApiClient} from '../getApiGatewayManagementApiClient'
import {getConnection} from '../getConnection'
import type {TrpcSubscriptionRequest} from './TrpcSubscriptionRequest'
import type {TrpcSubscriptionResponse} from './TrpcSubscriptionResponse'
import {isObservable} from './isObservable'

export const awsLambdaTrpcSubscriptionRequestHandler = <Router extends AnyTRPCRouter>(opts: {
  router: Router
  createContext?: (event: APIGatewayProxyEvent) => inferRouterContext<Router> | Promise<inferRouterContext<Router>>
}) => {
  const {router, createContext} = opts

  return async (event: APIGatewayProxyEvent, context: Context): Promise<APIGatewayProxyResult> => {
    context.callbackWaitsForEmptyEventLoop = false
    const connection = getConnection(event)

    if (!event.body || event.body === '[]') return {statusCode: 200, body: '{}'}

    const request = JSON.parse(event.body) as TrpcSubscriptionRequest

    const sendMessage = async (response: TrpcSubscriptionResponse) => {
      await getApiGatewayManagementApiClient(connection.endpoint).send(
        new PostToConnectionCommand({
          ConnectionId: connection.connectionId,
          Data: JSON.stringify(response),
        }),
      )
    }
    const isConnected = async (): Promise<boolean> => {
      return getApiGatewayManagementApiClient(connection.endpoint)
        .send(new GetConnectionCommand({ConnectionId: connection.connectionId}))
        .then(() => true)
        .catch(() => false)
    }
    const stopConnection = async (): Promise<void> => {
      await sendMessage({
        id: request.id,
        result: {type: 'stopped'},
      })
    }

    try {
      const messageQueue: TrpcSubscriptionResponse[] = []

      if (request.method === 'subscription.stop') {
        await stopConnection()
      }

      if (request.method === 'subscription') {
        messageQueue.push({
          id: request.id,
          result: {type: 'started'},
        })

        const procedure = await callTRPCProcedure({
          procedures: router._def.procedures,
          path: request.params.path,
          type: 'subscription',
          ctx: (await createContext?.(event)) || {},
          getRawInput: async () => request.params.input,
        })

        if (!isObservable(procedure)) {
          console.error(`observable`, procedure)
          throw `Procedure is not an observable!`
        }

        procedure.subscribe({
          next: (data: unknown) => {
            messageQueue.push({
              id: request.id,
              result: {
                type: 'data',
                data: data,
              },
            })
          },
          error: (err) => {
            console.error(err)
            messageQueue.push({
              id: request.id,
              result: {type: 'stopped'},
            })
          },
          complete: () => {
            messageQueue.push({
              id: request.id,
              result: {type: 'stopped'},
            })
          },
        })

        const pingIntervalMs = 5_000
        let lastPing = Date.now()
        let stopped = false
        while (!stopped) {
          if (messageQueue.length === 0) await new Promise((resolve) => setTimeout(resolve, 10))
          const message = messageQueue.shift()
          if (!message) continue

          await sendMessage(message)

          stopped = message.result.type === 'stopped'

          if (Date.now() > lastPing + pingIntervalMs) {
            if (!(await isConnected())) {
              console.info('Connection closed by client')
              stopped = true
            }
            lastPing = Date.now()
          }
        }
      }
    } catch (err) {
      if (err instanceof GoneException) {
        console.info('Connection closed by client')
      } else {
        console.error(err)
        await stopConnection()
      }
    }

    return {
      statusCode: 200,
      body: '{}',
    }
  }
}
